/**
 * Copyright 1998-2007, CHISEL Group, University of Victoria, Victoria, BC, Canada.
 * All rights reserved.
 */
package ca.uvic.csr.shrimp.gui.quickview;

import java.awt.Rectangle;
import java.awt.Window;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.Properties;
import java.util.Vector;

import javax.swing.Action;

import ca.uvic.csr.shrimp.LayoutConstants;
import ca.uvic.csr.shrimp.ShrimpProject.ShrimpProject;
import ca.uvic.csr.shrimp.util.ShrimpUtils;


/**
 * This class holds all the quick views.  It also provides methods for saving and loading
 * quick views from the project properties file.  
 * It also works with the project action manager to keep the quick view actions in sync.
 * 
 * @author Chris Callendar
 * @date 1-Aug-06
 */
public class QuickViewManager {

	// keys for saving quick views to a properties file
	private static final String PROPS_QUICKVIEWS_COUNT	= "quickviews.count";
	private static final String PROPS_QUICKVIEW 		= "quickview.";
	private static final String PROPS_NAME 				= ".name";
	private static final String PROPS_CLASS 			= ".class";
	private static final String PROPS_ICON 				= ".icon";
	private static final String PROPS_LABEL 			= ".label";
	private static final String PROPS_LAYOUT 			= ".layout";
	private static final String PROPS_NODES 			= ".nodes.count";
	private static final String PROPS_ALL_NODES 		= ".nodes.all";
	private static final String PROPS_NODE				= ".node.";
	private static final String PROPS_ARCS 				= ".arc.count";
	private static final String PROPS_ALL_ARCS 			= ".arcs.all";
	private static final String PROPS_ARC 				= ".arc.";
	private static final String PROPS_CPRELS_COUNT 		= ".cprels.count";
	private static final String PROPS_CPRELS_INVERTED 	= ".cprels.inverted";
	private static final String PROPS_CPREL 			= ".cprel.";
	private static final String PROPS_COMPOSITE_NODES_COUNT = ".composite.nodes.count";
	private static final String PROPS_COMPOSITE_NODE 	= ".composite.node.";
	private static final String PROPS_GROUPS 			= ".composite.groups.count";
	private static final String PROPS_GROUP 			= ".composite.group.";
	private static final String PROPS_GROUP_ARCS 		= ".arcs.count";
	private static final String PROPS_GROUP_ARC 		= ".arc.";
	private static final String PROPS_BOUNDS			= "quickviews.bounds";
	
	// quick view actions
	private Collection quickViews;	
	private ShrimpProject project;
	
	/**
	 * Initializes the manager.
	 */
	public QuickViewManager(ShrimpProject project) {
		this.project = project;
		this.quickViews = new ArrayList();
	}
	
	/**
	 * Gets all the quick view actions.
	 * @return the {@link Action}s
	 */
	public Collection getQuickViews() {
		return quickViews;
	}
	
	public ShrimpProject getProject() {
		return project;
	}
	
	/**
	 * Returns a sorted array of quick views.
	 * @see QuickViewComparator
	 */
	public QuickViewAction[] getSortedQuickViews() {
		Collection actions = getQuickViews();
		QuickViewAction[] quickViewsArray = (QuickViewAction[])actions.toArray(new QuickViewAction[actions.size()]);
		Arrays.sort(quickViewsArray, new QuickViewComparator());
		return quickViewsArray;
	}
		
	/**
	 * Gets a quick view action with the given name.
	 * @param name
	 * @return {@link QuickViewAction} or null if not found
	 */
	public QuickViewAction getQuickViewAction(String name) {
		QuickViewAction action = null;
		for (Iterator iter = getQuickViews().iterator(); iter.hasNext(); ) {
			QuickViewAction qva = (QuickViewAction) iter.next();
			if (qva.getActionName().equals(name)) {
				action = qva;
				break;
			}
		}
		return action;
	}
	
	/**
	 * Removes a quick view with the given name.
	 * @param name
	 * @return QuickViewAction or null if it doesn't exist.
	 */
	public QuickViewAction removeQuickView(String name) {
		QuickViewAction action = getQuickViewAction(name);
		if (action != null) {
			quickViews.remove(action);
		}
		return action;
	}
	
	/**
	 * Removes a quick view action from the collection.
	 * @param action
	 * @return Action or null if it doesn't exist.
	 */
	public boolean removeQuickView(QuickViewAction action) {
		boolean removed = false;
		if (action != null) {
			removed = quickViews.remove(action);
		}
		return removed;
	}
	
	/**
	 * Adds a quick view to the collection.  
	 * @param action
	 */
	public void addQuickView(QuickViewAction action) {
		if ((action != null) && !quickViews.contains(action)) {
			quickViews.add(action);
		}
	}
	
	/**
	 * Checks if that action exists in the quick views collection.
	 * @param action
	 * @return true if the action is found
	 */
	public boolean hasQuickView(QuickViewAction action) {
		return quickViews.contains(action);
	}
	
	/**
	 * Checks if a quickview with that name exists.
	 * @param name
	 * @return true if an action with that name exists
	 */
	public boolean hasQuickView(String name) {
		return (getQuickViewAction(name) != null);
	}	
	
	/**
	 * Saves the quicks views out to the project properties file.
	 */
	public void save() {
		if (project != null) {
			Properties props = project.getProperties();
			
			// remove the quick view properties
			removeAllQuickViewPropertes(props);
			
			// now put the quick view properties back in
			Collection qvs = getQuickViews();
			props.setProperty(PROPS_QUICKVIEWS_COUNT, String.valueOf(qvs.size()));
			int index = 0, allNodes = -1, allArcs = -1;
			for (Iterator iter = qvs.iterator(); iter.hasNext(); ) { 
				QuickViewAction action = (QuickViewAction) iter.next();
				if (allNodes == -1) {
					allNodes = action.getAllNodeTypes().size();
					allArcs = action.getAllArcTypes().size();
				}
				saveQuickView(props, action, index, allNodes, allArcs);
				index++;
			}
			
			// force writing out to file
			project.saveProperties();
		}
	}
	
	/**
	 * Saves the bounds in the project properties file.
	 * @param window
	 */
	public void saveBounds(Window window) {
		if ((window != null) && (project != null)) {
			Properties props = project.getProperties();
			Rectangle bounds = window.getBounds();
			String str = bounds.x+","+bounds.y+","+bounds.width+","+bounds.height;
			props.setProperty(PROPS_BOUNDS, str);
		}
	}
	
	/**
	 * Loads the bounds from the project properties file.
	 * @param window
	 */
	public void loadBounds(Window window) {
		if ((window != null) && (project != null)) {
			Properties props = project.getProperties();
			String str = props.getProperty(PROPS_BOUNDS, null);
			boolean ok = false;
			if (str != null) {
				String[] xywh = str.split(",");
				if (xywh.length == 4) {
					try {
						Rectangle rect = new Rectangle(
								Math.max(0,Integer.parseInt(xywh[0])),		// X
								Math.max(0, Integer.parseInt(xywh[1])),		// Y
								Math.max(100, Integer.parseInt(xywh[2])),	// WIDTH
								Math.max(100, Integer.parseInt(xywh[3])));	// HEIGHT
						window.setBounds(rect);
						ok = true;
					} catch (Exception ignore) {}
				}
			}			
			if (!ok) {
				// center on parent
				ShrimpUtils.centerWindowOnParent(window, window.getParent());
			}
		}
	}
	
	/**
	 * Removes all the properties that start with the String <code>quickview.</code>
	 */
	private void removeAllQuickViewPropertes(Properties props) {
		Enumeration keys = props.propertyNames();
		Collection toRemove = new LinkedList();
		while (keys.hasMoreElements()) {
			String key = (String) keys.nextElement();
			if (key.startsWith(PROPS_QUICKVIEW)) {
				toRemove.add(key);
			}
		}
		for (Iterator iter = toRemove.iterator(); iter.hasNext(); ) {
			String key = (String) iter.next();
			props.remove(key);
		}
	}

	/**
	 * Sets the properties for a quick view.
	 * @param index	the index of the quick view (starting at 0)
	 * @param allNodes the total number of node types
	 * @param allArcs the total number of arc types
	 */
	private void saveQuickView(Properties props, QuickViewAction action, int index, int allNodes, int allArcs) {
		String key = PROPS_QUICKVIEW + index;
		props.setProperty(key + PROPS_CLASS, action.getClass().getName());
		props.setProperty(key + PROPS_NAME, action.getActionName());
		props.setProperty(key + PROPS_ICON, action.getIconFilename());
		ExpressViewConfigurator config = action.getConfigurator();
		props.setProperty(key + PROPS_LAYOUT, config.getLayoutMode());
		props.setProperty(key + PROPS_LABEL, config.getLabelMode());
		
		Collection nodes = config.getNodeTypesOfInterest();
		if (nodes.size() == allNodes) {
			// no point in saving all of the node types 
			props.setProperty(key + PROPS_ALL_NODES, "true");
		} else {
			props.setProperty(key + PROPS_NODES, String.valueOf(nodes.size()));
			int typeIndex = 0;
			for (Iterator iter = nodes.iterator(); iter.hasNext(); ) {
				String type = (String) iter.next();
				props.setProperty(key + PROPS_NODE + typeIndex, type);
				typeIndex++;
			}
		}

		Collection arcs = config.getArcTypesOfInterest();
		if (arcs.size() == allArcs) {
			props.setProperty(key + PROPS_ALL_ARCS, "true");
		} else {
			props.setProperty(key + PROPS_ARCS, String.valueOf(arcs.size()));
			int typeIndex = 0;
			for (Iterator iter = arcs.iterator(); iter.hasNext(); ) {
				String type = (String) iter.next();
				props.setProperty(key + PROPS_ARC + typeIndex, type);
				typeIndex++;
			}
		}
		
		String[] cprels = config.getCprels();
		props.setProperty(key + PROPS_CPRELS_COUNT, String.valueOf(cprels.length));
		props.setProperty(key + PROPS_CPRELS_INVERTED, String.valueOf(config.isInverted()));
		for (int i = 0; i < cprels.length; i++) {
			props.setProperty(key + PROPS_CPREL + i, cprels[i]);
		}
		
		// save composite nodes to open
		Collection nodesToOpen = config.getNodeTypesToOpen();
		props.setProperty(key + PROPS_COMPOSITE_NODES_COUNT, String.valueOf(nodesToOpen.size()));
		index = 0;
		for (Iterator iter = nodesToOpen.iterator(); iter.hasNext(); ) {
			String type = (String) iter.next();
			props.setProperty(key + PROPS_COMPOSITE_NODE + index, type);
			index++;
		}
		
		// save composite arc types and groups
		Map groups = config.getCompositeArcs();
		props.setProperty(key + PROPS_GROUPS, String.valueOf(groups.size()));
		int groupIndex = 0;
		for (Iterator iter = groups.keySet().iterator(); iter.hasNext(); ) {
			String groupName = (String) iter.next();
			String key2 = key + PROPS_GROUP + groupIndex;
			props.setProperty(key2, groupName);
			Collection cArcs = (Collection) groups.get(groupName);
			props.setProperty(key2 + PROPS_GROUP_ARCS, String.valueOf(cArcs.size()));
			int typeIndex = 0;
			for (Iterator iter2 = cArcs.iterator(); iter2.hasNext(); ) {
				String arcType = (String) iter2.next();
				props.setProperty(key2 + PROPS_GROUP_ARC + typeIndex, arcType);
				typeIndex++;
			}
			groupIndex++;
		}
	}

	/**
	 * Loads the quick views from the properties file.
	 * @param clearQuickViews if true the current quick views will be cleared
	 * @return true if more than one property was loaded
	 */
	public boolean loadQuickViewsFromProperties(boolean clearQuickViews) {
		if (clearQuickViews) {
			quickViews.clear();
		}
		int loadedCount = 0;
		if (project != null) {
			Properties props = project.getProperties();
			int count = Integer.parseInt(props.getProperty(PROPS_QUICKVIEWS_COUNT, "0"));
			for (int i = 0; i < count; i++) {
				QuickViewAction action = loadQuickView(props, i);
				if (action != null) {
					loadedCount++;
					addQuickView(action);
				}
			}
		}
		return (loadedCount > 0);
	}	

	/**
	 * Loads the quick view from the properties.  Creates the {@link QuickViewAction} class
	 * and sets its properties.
	 * @param index the index of the quick view (starting at 0)
	 * @return QuickViewAction or null
	 */
	private QuickViewAction loadQuickView(Properties props, int index) {
		QuickViewAction action = null;
		String baseKey = PROPS_QUICKVIEW + index;
		String key = baseKey + PROPS_CLASS;
		try {
			String clsName = props.getProperty(key);
			// create the class
			try {
				Class cls = Class.forName(clsName);
				action = (QuickViewAction) cls.newInstance();
			} catch (Throwable t) {
				System.err.println("Warning - couldn't create a quick view of type " + clsName);
				action = new QuickViewAction();
			}
			
			action.setProject(project);
			key = baseKey + PROPS_NAME;
			action.putValue(Action.NAME, props.getProperty(key));
			key = baseKey + PROPS_ICON;
			action.setIconFilename(props.getProperty(key, "icon_quick_view_blank.gif"));
			ExpressViewConfigurator config = action.getConfigurator();
			key = baseKey + PROPS_LAYOUT;
			config.setLayoutMode(props.getProperty(key, LayoutConstants.LAYOUT_GRID_BY_TYPE));
			key = baseKey + PROPS_LABEL;
			config.setLabelMode(props.getProperty(key, QuickViews.DEFAULT_LABEL_MODE));
	
			Collection nodeTypes;
			key = baseKey + PROPS_ALL_NODES;
			boolean all = Boolean.parseBoolean(props.getProperty(key, "false"));
			if (all) {
				// all the node types were chosen
				nodeTypes = action.getAllNodeTypes();
			} else {
				key = baseKey + PROPS_NODES;
				int count = Integer.parseInt(props.getProperty(key, "0"));
				nodeTypes = new HashSet(count);
				for (int i = 0; i < count; i++) {
					key = baseKey + PROPS_NODE + i;
					String type = props.getProperty(key, null);
					if (type != null) {
						nodeTypes.add(type);
					}
				}
			}
			config.setNodeTypesOfInterest(nodeTypes);
	
			
			Collection arcTypes;
			key = baseKey + PROPS_ALL_ARCS;
			all = Boolean.parseBoolean(props.getProperty(key, "false"));
			if (all) {
				arcTypes = action.getAllArcTypes();
			} else {
				key = baseKey + PROPS_ARCS;
				int count = Integer.parseInt(props.getProperty(key, "0"));
				arcTypes = new HashSet(count);
				for (int i = 0; i < count; i++) {
					key = baseKey + PROPS_ARC + i;
					String type = props.getProperty(key, null);
					if (type != null) {
						arcTypes.add(type);
					}
				}
			}
			config.setArcTypesOfInterest(arcTypes);
			
			key = baseKey + PROPS_CPRELS_COUNT;
			int count = Integer.parseInt(props.getProperty(key, "0"));
			key = baseKey + PROPS_CPRELS_INVERTED;
			config.setInverted(Boolean.parseBoolean(props.getProperty(key, "false")));
			String[] cprels = new String[count];
			for (int i = 0; i < count; i++) {
				key = baseKey + PROPS_CPREL + i;
				cprels[i] = props.getProperty(key, null);
			}
			config.setCprels(cprels);
			
			// load composite nodes to open
			key = baseKey + PROPS_COMPOSITE_NODES_COUNT;
			count = Integer.parseInt(props.getProperty(key, "0"));
			Collection nodesToOpen = new Vector(count);
			for (int i = 0; i < count; i++) {
				key = baseKey + PROPS_COMPOSITE_NODE + i;
				String type = props.getProperty(key, null);
				if (type != null) {
					nodesToOpen.add(type);
				}
			}
			config.setNodeTypesToOpen(nodesToOpen);
			
			// load composite arc types and groups
			key = baseKey + PROPS_GROUPS;
			count = Integer.parseInt(props.getProperty(key, "0"));
			HashMap groups = new HashMap(count);
			for (int i = 0; i < count; i++) {
				key = baseKey + PROPS_GROUP + i;
				String groupName = props.getProperty(key, null);
				if (groupName != null) {
					String groupKey = key;
					key = groupKey + PROPS_GROUP_ARCS;
					int arcCount = Integer.parseInt(props.getProperty(key, "0"));
					Collection cArcs = new Vector(arcCount);
					for (int j = 0; j < arcCount; j++) {
						key = groupKey + PROPS_GROUP_ARC + j;
						String arcType = props.getProperty(key, null);
						if (arcType != null) {
							cArcs.add(arcType);
						}
					}
					groups.put(groupName, cArcs);
				}
			}
			config.setCompositeArcs(groups);
		} catch (Exception e) {
			System.err.println("Error loading quick views.  Last property key was '" + key + "'.");
		}
		
		return action;
	}

	/**
	 * Clears the current quick views and then loads the default quick views for the given project.
	 */
	public void loadDefaults() {
		quickViews.clear();
		if (project != null) {
			project.createDefaultQuickViewActions();
		}
	}

	/**
	 * Cleanup.
	 */
	public void dispose() {
		for (Iterator iter = quickViews.iterator(); iter.hasNext(); ) {
			QuickViewAction action = (QuickViewAction) iter.next();
			action.dispose();
		}
		quickViews = new ArrayList(1);
		project = null;
	}
	
}
